#include <iostream>
#include <string>
#include <getopt.h>
#include "systance/window.hh"
#include "systance/time.hh"
#include "systance/frame_buffer.hh"
#include "loadance/loadance.hh"
#include "maintance/base/scene.hh"
#include "maintance/raytrace/raytracer.hh"
#include "config.h"

std::string program_name = "";
std::string xmlfile = "";

enum exit_codes
  {
    exit_ok = 0,
    loading_error = 1,
    init_error = 2,
    unknown_error = 9
  };

int exit_status = exit_ok;

static void quit_program ()
{
  switch (exit_status)
    {
    case exit_ok:
      break;
    case loading_error:
      std::cerr << program_name << ": loading error" << std::endl;
      break;
    case init_error:
      std::cerr << program_name << ": error while making init" << std::endl;
      break;
    case unknown_error:
    default:
      std::cerr << program_name << ": unknown error" << std::endl;
      break;
    }
  if (exit_status != exit_ok)
    std::cerr << "/!\\ exiting /!\\" << std::endl;
  exit (exit_status);
}

// getopt part
int		verbose = 0;
unsigned int	width = 320;
unsigned int	height = 240;
unsigned int	bpp = 16;
unsigned int	base = 4;
unsigned int	pixel = 1;
bool		fullscreen = false;
bool		wire = false;
bool		benchmark = false;

enum short_options_e
  {
    opt_help = 'h',
    opt_version = 'V',
    opt_verbose = 'v',
    opt_res = 'r',
    opt_screen = 'f',
    opt_bpp = 't',
    opt_pixel = 'p',
    opt_wire = 'w',
    opt_base = 'a',
    opt_benchmark = 'b'
  };

static struct option const long_options[] =
  {
    {"help",            no_argument,    0,		opt_help},
    {"version",		no_argument,	0,		opt_version},
    {"verbose",         no_argument,    &verbose,	1},
    {"res",       required_argument,    0,		opt_res},
    {"fullscreen",	no_argument,    0,		opt_screen},
    {"bpp",	  required_argument,	0,		opt_bpp},
    {"pixel",	  required_argument,	0,		opt_pixel},
    {"wireframe",       no_argument,	0,		opt_wire},
    {"base",	  required_argument,	0,		opt_base},
    {"benchmark",       no_argument,	0,		opt_benchmark},
    {NULL, 0, NULL, 0}
  };

static const char short_options[] = "hvVfwb";

// general fonctions
static void version (void)
{
  std::cerr << PACKAGE << ", " << VERSION << std::endl;
  std::cerr << "@2002-2003 Alexandre Chouippe, Laurent Corbes, Henry Precheur, Olivier Leclant" << std::endl;
  std::cerr << "Release under the terms of the X11 License" << std::endl;
}

static void usage (void)
{
  version();
  std::cerr << "usage: " << program_name << " [-h|-V|-v|-f|-w|-b] "
	    << "xmlfile" << std::endl;
  std::cerr << "..." << std::endl
	    << "-h      --help		       print this message" 
	    << std::endl
	    << "-V      --version	       print program version" 
	    << std::endl
	    << "-v      --verbose	       print debug messages"
	    << std::endl
	    << "        --res=valxval          select widthxheight"
	    << std::endl
	    << "-f      --fullscreen           turn on fullscreen mode"
	    << std::endl
	    << "        --bpp=val              select depth of color, value are 16 or 32"
	    << std::endl
	    << "        --pixel=val            select pixel width (a big value makes "
	    << std::endl
	    << "                               the rendering faster and uglier)"
	    << std::endl
	    << "-w      --wireframe            use wireframe rendering (for debuging)"
	    << std::endl
	    << "        --base=val             define the size of base scare"
	    << std::endl
	    << "-b      --benchmark            run the scene the faster possible and then print"
	    << std::endl
	    << "                               the average fps"
	    << std::endl;
  quit_program ();
}

// main fonction just do the getopt and set the correct runtime options
int main (int argc, char **argv)
{
  program_name = argv[0];
  
  int c;
  
  while ((c = getopt_long (argc, argv, short_options, long_options, NULL))
         != -1)
    {
      switch (c)
        {
        case opt_help:
          usage();
          break;
	case opt_version:
	  version();
	  quit_program ();
	  break;
	case opt_verbose:
	  verbose = 1;
	  break;
	case opt_res:
	  width = strtol(strtok(optarg, "x"), NULL, 10);
	  height = strtol(strtok(NULL, "\0"), NULL, 10);
	  break;
	case opt_screen:
	  fullscreen = true;
	  break;
	case opt_bpp:
	  bpp = strtol(optarg, NULL, 10);
	  break;
	case opt_pixel:
	  pixel = strtol(optarg, NULL, 10);
	  break;
	case opt_wire:
	  wire = true;
	  break;
	case opt_base:
	  base = strtol(optarg, NULL, 10);
	  break;
	case opt_benchmark:
	  benchmark = true;
	  break;
        }
    }
  if (optind != argc - 1)
    usage ();
  
  xmlfile = argv[optind];

  // loading xml file
  loadance::LoadXml xmlscene(xmlfile);
  maintance::Scene* scene = xmlscene.load();
  if (!scene)
    {
      exit_status = loading_error;
      quit_program();
    }
  scene->SetTime(0.0f);
  scene->print(std::cout);
  
  // Core system
  systance::Time	time;
  systance::UnixWindow	win;
  systance::FrameBuffer	fb;
  
  if ((!win.Create(width, height, bpp, fullscreen)) ||
      (!(fb.Create(wire, pixel, width, height))))
    {
      exit_status = init_error;
      quit_program();
    }
  raytrace::Raytracer rayt(*scene, fb.GetWidth(), fb.GetHeight());
  // fb.SetBase(base);
  fb.Attach(rayt);
  time.Init();

  unsigned int	frame_draw = 0;

  if (!benchmark)
    {
      // le temps de la demo avant de boucler
      unsigned int maxtime = 10;

      char	title[128];
      float	time_last = -1;

      while(!win.IsQuit())
	{
	  time.Lock();
	  if (time.Get() - time_last >= 1.0f)
	    {
	      sprintf(title, PACKAGE " - " VERSION " %u fps, %u rayz %f sec",
		      frame_draw, fb.GetRayCount(), time.Get());
	      win.SetWindowTitle(title);
	      frame_draw = 0;
	      time_last = time.Get();
	    }
	  else
	    frame_draw++;
	  scene->SetTime(time.Get()*24.0f);
	  rayt.Update();
	  fb.Trace();
	  
	  win.Draw();
	  if (time.Get() >= maxtime)
 	    {
 	      time.Init();
 	      frame_draw = 0;
 	      time_last = time.Get();
 	    }
 	  else
	    time.UnLock();
	}
    }
  else
    {
      float stime = time.Get();
      unsigned int nbframe = 500;
      win.SetWindowTitle("benchmarkmode");
      while(frame_draw <= nbframe)
	{
	  frame_draw++;
	  scene->SetTime(frame_draw);
	  rayt.Update();
	  fb.Trace();
	  win.Draw();
	}
      float etime = time.Get();
      std::cout << "---- time = " << etime - stime << "sec ----" << std::endl;
      std::cout << "---- fps = " << nbframe / (etime - stime) << " ----" << std::endl;
    }

  win.Destroy();
  delete scene;
  quit_program ();
}
