#ifndef MATHSLIB_HH_
#define MATHSLIB_HH_
#include <cmath>
#include <cassert>

/*---------------------------------------------------------------------------*\
|Mise en cache des valeurs pour certaines fonctions mathmatiques standards   |
|Or de l'intervale, si la fonction n'est pas periodique l'appel a la          |
|fonction de base est utilise. Dans l'intervale, on utilise des interpolations|
lineaires entre les echantillons                                              |
\*----------------------------------------------------------------------------*/
#ifndef M_PI
#	define M_PI       3.14159265358979323846
#endif

#define PI M_PI
#define TO_RAD(x) (PI/180 * x)
namespace math
{

template<typename T>
class DefaultFunction
{
public :
  typedef T (*MathFct)(T x);

  /*------------------------------------------------------------------
    MathFct fct : fonction mathematique a utiliser (sin, cos, log ...)
    T begin, T end : Intervale de calcul du cache
    T step : Taille de l'echantillion
    bool cyclic : fonction periodique sur l'intervalle
    -----------------------------------------------------------------*/

  DefaultFunction(MathFct fct, T begin, T end, T step, bool cyclic = false)
  {
    assert(step != 0 && fct != 0);
    assert(begin < end);
    _fct = fct;
    _map = 0;
    _begin = begin;
    _end = end;
    _step = step;
    _cyclic = cyclic;
    init_map();
  }

  DefaultFunction()
  { 
    _map = 0;
    _begin = 0;
    _end = 0;
    _step = 0;
    _cyclic = 0;
    _size = 0;
    _map = 0;
  }


  ~DefaultFunction()
  {
    if (_map)
      delete _map;
  }


  inline T operator () (T x)
  {
    return value(x);
  }


  T value(T x)
  {
    if (_cyclic)
      {
	T period = _end - _begin;
	while (x > _end)
	  x -= period;
	while (x < _begin)
	  x += period;
      }
    
    if (!_map || x > _end || x < _begin)
      return _fct(x);
    
    int index = get_index(x);
    return _map[index].y + 
      ((x - _map[index].x) / ( _map[index + 1].x - _map[index].x)) * (_map[index + 1].y - _map[index].y);
    
    return (0);
  }
  
private :

  void init_map()
  {
    _size = (int) ((_end - _begin) / _step + 1);
    _map = new Val[_size];
    
    int i;
    T j;

    for ( i = 0, j = _begin ; i < _size; i++, j += _step)
      {
	_map[i].x = j;
	_map[i].y = _fct(j);
      }
  }

  int get_index(T x) 
  {
    x = x - _begin;
    return (int) (x/_step);
  }

  // Utilise pour stocker la table de cache
  typedef struct s_val
  {
    T x;
    T y;
  } Val;

  T _begin;
  T _end;
  T _step;
  
  int _size;
  bool _cyclic;
  
  Val* _map;
  MathFct _fct;

};

  /*-----------------------------
    Fonctions utilisant un cache
    -----------------------------*/
  
/*  static float fsin(float f) {return sin(f);}
  static float fcos(float f) {return cos(f);}
  static float ftan(float f) {return tan(f);}
  
  static float fsqrt(float f) {return sqrt(f);}
  
  
  static DefaultFunction<float> Sin(fsin, 0, (float)(2 * PI), 0.1f, true);
  static DefaultFunction<float> Cos(fcos, 0, (float)(2 * PI), 0.1f, true);
  static DefaultFunction<float> Tan(ftan, 0, (float)(2 * PI), 0.1f, false);
  static DefaultFunction<float> Sqrt(fsqrt, 0, 500, 0.5f);
*/  
}

#endif // MATHSLIB_HH_
