#ifndef _TYPES_MATRIX_HH_
#define _TYPES_MATRIX_HH_

#include <cassert>
#include <cmath>
#include "streamable.hh"

// classe generique Matrix<T>
// se construit de la forme Matrix (m, n)
// elle est de la forme:
//    - n -  
// |{       }
// m{       }
// |{       }
// comme les matrices francaises
//
// operations et fonctions implementes:
// 
// implemente streamable (print et read)
// Matrix<T> + Matrix<T>
// Matrix<T> += Matrix<T>
// Matrix<T> - Matrix<T>
// Matrix<T> -= Matrix<T>
// Matrix<T> * Matrix<T>
// Matrix<T> * T
// T * Matrix<T>
// Matrix<T> / T
// T / Matrix<T>
// idmult (retourne une matrice qui est la mult id de celle en cours)
// transpose (retourne une Matrix qui est la transpose de celle en cours)
// trace (retourne la trace de cette matrix)
// norme (retrourne la norme de la matrix courante)
// copy (retourne dans une nouvelle Matrix une partie de la matrice de depart
//       les deux premiers param sont les points de depart
//       les deux autres sont les points d'arrive)
// val_propre (retourne la plus grande val propore de la Matrix)
// vect_propre (retourne le vecteur propre de la plus grande val propre
//              sous forme d'une matrix)
// 
// les accesseurs sont (m, n)

namespace types
{
	template <typename T>
	class Matrix: public Streamable
	{
	public:
		Matrix (int m, int n): _m (m), _n (n), _nbelt (m * n)
		{ 
			_tab = new T[_nbelt]; 
			assert (_tab != 0);
			for (int i = 0; i < _nbelt; ++i)
				_tab[i] = 0;
		}
		Matrix (const Matrix& rhs): Streamable(), _m (rhs._m), _n (rhs._n), _nbelt (rhs._nbelt)
		{
			_tab = new T[_nbelt]; 
			assert (_tab != 0);
			for (int i = 0; i < _nbelt; ++i)
				_tab[i] = rhs._tab[i];
		}
		Matrix& operator= (const Matrix& rhs)
		{
			if (&rhs == this) return *this;
			_m = rhs._m; _n = rhs._n; _nbelt = rhs._nbelt;
			_tab = new T[_nbelt]; 
			assert (_tab != 0);
			for (int i = 0; i < _nbelt; ++i)
				_tab[i] = rhs._tab[i];
			return *this;
		}
		virtual ~Matrix () 
		{ delete[] _tab; }

		void print(std::ostream& os) const 
		{ 
			for (int i = 1; i <= _m; ++i)
			{
				for (int j = 1; j <= _n; ++j)
					os << (*this)(i, j) << " ";
				os << std::endl;
			}
			os << std::endl;
		}
		void read(std::istream& is) 
		{
			for (int i = 0; i < _nbelt; ++i)
				is >> _tab[i] ;
		}

		Matrix<T> operator+ (const Matrix& rhs) const
		{
			assert ((_m == rhs._m) && (_n == rhs._n));
			Matrix<T> tmp (_m, _n);
			for (int i = 0; i < _nbelt; ++i)
				tmp._tab[i] = _tab[i] + rhs._tab[i];
			return tmp;
		}
		Matrix<T>& operator+= (const Matrix& rhs)
		{
			assert ((_m == rhs._m) && (_n == rhs._n));
			for (int i = 0; i < _nbelt; ++i)
				_tab[i] += rhs._tab[i];
			return *this;
		}
		Matrix<T> operator- (const Matrix& rhs) const
		{
			assert ((_m == rhs._m) && (_n == rhs._n));
			Matrix<T> tmp (_m, _n);
			for (int i = 0; i < _nbelt; ++i)
				tmp._tab[i] = _tab[i] - rhs._tab[i];
			return tmp;
		}
		Matrix<T>& operator-= (const Matrix& rhs)
		{
			assert ((_m == rhs._m) && (_n == rhs._n));
			for (int i = 0; i < _nbelt; ++i)
				_tab[i] -= rhs._tab[i];
			return *this;
		}
		Matrix<T> operator* (const Matrix& rhs) const
		{
			assert (_n == rhs._m);
			Matrix<T> tmp (_m, rhs._n);
			for (int i = 1; i <= tmp._m; ++i)
				for (int j = 1; j <= tmp._n; ++j)
					for (int k = 1; k <= _n; ++k)
						tmp(i, j) += (*this)(i, k) * rhs(k, j);
			return tmp;
		}
		Matrix<T> operator* (const T& rhs) const
		{
			Matrix<T> tmp (_m, _n);
			for (int i = 1; i <= tmp._m; ++i)
				for (int j = 1; j <= tmp._n; ++j)
					tmp(i, j) = (*this)(i, k) * rhs;
			return tmp;
		}
		friend Matrix<T> operator* (const T& lhs, const Matrix& rhs)
		{
			Matrix<T> tmp (rhs._m, rhs._n);
			for (int i = 1; i <= tmp._m; ++i)
				for (int j = 1; j <= tmp._n; ++j)
					tmp(i, j) = rhs(i, j) * lhs;
			return tmp;
		}
		Matrix<T> operator/ (const T& rhs) const
		{
			assert (rhs != 0);
			Matrix<T> tmp (_m, _n);
			for (int i = 1; i <= tmp._m; ++i)
				for (int j = 1; j <= tmp._n; ++j)
					tmp(i, j) = (*this)(i, j) / rhs;
			return tmp;
		}
		friend Matrix<T> operator/ (const T& lhs, const Matrix& rhs)
		{
			assert (lhs != 0);
			Matrix<T> tmp (rhs._m, rhs._n);
			for (int i = 1; i <= tmp._m; ++i)
				for (int j = 1; j <= tmp._n; ++j)
					tmp(i, j) = rhs(i, j) / lhs;
			return tmp;
		}

		Matrix<T> idmult ()
		{
			assert (_m == _n);
			Matrix<T> id (_m, _n);
			for (int i = 1; i <= _m; ++i)
				id(i, i) = 1;
			Matrix<T> tmp (_m, _n);
			for (int i = 1; i <= _m; ++i)
				for (int j = 1; j <= _n; ++j)
					for (int k = 1; k <= _n; ++k)
						tmp(i, j) += (*this)(i, k) * id(k, j);
			return tmp;
		}
		Matrix<T> transpose () const
		{
			assert (_m == _n);
			Matrix<T> tmp (_m, _n);
			for (int i = 1; i <= _m; ++i)
				for (int j = 1; j <= _n; ++j)
					tmp(i, j) = (*this)(j, i);
			return tmp;
		}
		T trace () const
		{
			assert (_m == _n);
			T res = 0;
			for (int i = 1; i <= _m; ++i)
				res += (*this)(i, i);
			return res;
		}
		T norme () const
		{
			T tmp = 0;
			for (int i = 1; i <= _m; ++i)
				for (int j = 1; j <= _n; ++j)
					tmp += pow (((*this)(i, j)), 2);
			return sqrt (res);
		}
		Matrix<T> copy (const int x, const int y, const int m, const int n) const
		{
			assert ((x >= 1) && (x <= _m) && (y >= 1) && (y <= _n));
			assert((n >= x) && (m >= y));
			Matrix<T> tmp (m - x + 1, n - y + 1);
			for (int i = 1; i <= tmp._m; ++i)
				for (int j = 1; j <= tmp._n; ++j)
					tmp(i, j) = (*this)(x + i - 1, y + j - 1);
			return tmp;
		}
		T val_propre () const
		{
			assert (_m == _n);
			const T eps = 0.0001;
			T l_old = 0, l = 0;
			Matrix<T> tmp (_m, 1);

			for (int i = 1; i <= _m; ++i)
				tmp (i, 1) = 1;
			Matrix<T> Atmp (_m, 1);
			do 
			{
				for (int i = 1; i <= _m; ++i)
				{
					Matrix<T> atmpj (this->copy(i, 1, i, _n));
					Atmp (i, 1) = (atmpj * tmp)(1,1);
				}
				if (l != 0)
					l_old = l;
				l = Atmp (1,1);
				for (int i = 1; i <= _m; ++i)
					tmp(i, 1) = Atmp(i, 1) / l;
			}
			while (fabs (l - l_old) >= eps);
			return l;
		}
		Matrix<T> vect_propre () const
		{
			assert (_m == _n);
			const T eps = 0.0001;
			T l_old = 0, l = 0;
			Matrix<T> tmp (_m, 1);

			for (int i = 1; i <= _m; ++i)
				tmp (i, 1) = 1;
			Matrix<T> Atmp (_m, 1);
			do 
			{
				for (int i = 1; i <= _m; ++i)
				{
					Matrix<T> atmpj (this->copy(i, 1, i, _n));
					Atmp (i, 1) = (atmpj * tmp)(1,1);
				}
				if (l != 0)
					l_old = l;
				l = Atmp (1,1);
				for (int i = 1; i <= _m; ++i)
					tmp(i, 1) = Atmp(i, 1) / l;
			}
			while (fabs (l - l_old) >= eps);
			return tmp;
		}

		T operator()(const int m, const int n) const
		{
			assert ((m >= 1) && (m <= _m));
			assert ((n >= 1) && (n <= _n));
			return (_tab[(m - 1) * _n + n - 1]);
		}
		T& operator()(const int m, const int n)
		{
			assert ((m >= 1) && (m <= _m));
			assert ((n >= 1) && (n <= _n));
			return (_tab[(m -1) * _n + n - 1]);
		}

		int getm () const { return _m; }
		int getn () const { return _n; }

	private:
		T* _tab;
		int _m;
		int _n;
		int _nbelt;
	};
}

#endif // TYPES_MATRIX_HH
