#ifndef _TYPES_QUATERNION_HH_
#define _TYPES_QUATERNION_HH_

#include <cmath>
#include "streamable.hh"

// classe generique Quaternion<T>
// se construit Quaternion (w, x, y, z)
//
// operations et fonctions implementes:
//
// implemente streamable (print et read)
// Quaternion + Quaternion
// Quaternion += Quaternion
// Quaternion - Quaternion
// Quaternion -= Quaternion
// Quaternion * Quaternion
// Quaternion *= Quaternion
// Quaternion * T
// Quaternion *= T
// double * Quaternion
// Quaternion / Quaternion
// Quaternion /= Quaternion
// Quaternion / T
// ~Quaternion (conjugue)
// get_norm (retourne la norme)
// normalise (normalise le quaternion)
// conjugate (retourne un Quaternion)
// Slerp q1 q2 step
// les accesseurs get_w()... et set_x(T)...

namespace types
{
	template <typename T>
	class Quaternion: public Streamable
	{
	public:
		Quaternion (): _w (0), _x (0),_y (0),_z (0) {}
		Quaternion (const T w, const T x, const T y, const T z)
			: _w (w), _x (x),_y (y),_z (z) {}
			Quaternion (const Quaternion& rhs)
				: Streamable(), _w (rhs._w), _x (rhs._x), _y (rhs._y), _z (rhs._z) {}
				Quaternion& operator= (const Quaternion& rhs) 
				{
					if (&rhs == this)	return *this;
					_w = (rhs._w); _x = rhs._x; _y = rhs._y; _z = rhs._z; 
					return *this;
				}
				virtual ~Quaternion () {}

				void print(std::ostream& os) const 
				{ os << "(" << _w << " " << _x << " " << _y << " " << _z << ")"; }
				void read(std::istream& is) 
				{ is >> _w >> _x >> _y >> _z; }

				Quaternion operator+(const Quaternion& rhs) const
				{
					Quaternion tmp (_w + rhs._w, _x + rhs._x, _y + rhs._y, _z + rhs._z);
					return tmp;
				}
				Quaternion& operator+=(const Quaternion& rhs)
				{
					_w += rhs._w; _x += rhs._x; _y += rhs._y; _z += rhs._z;
					return *this;
				}
				Quaternion operator-(const Quaternion& rhs) const
				{
					Quaternion tmp (_w - rhs._w, _x - rhs._x, _y - rhs._y, _z - rhs._z);
					return tmp;
				}
				Quaternion& operator-=(const Quaternion& rhs)
				{
					_w -= rhs._w; _x -= rhs._x; _y -= rhs._y; _z -= rhs._z;
					return *this;
				}
				Quaternion operator*(const Quaternion& rhs) const
				{
					Quaternion tmp (_w * rhs._w - _x * rhs._x - 
						_y * rhs._y - _z * rhs._z,
						_w * rhs._x + _x * rhs._w + 
						_y * rhs._z - _z * rhs._y,
						_w * rhs._y - _x * rhs._z + 
						_y * rhs._w + _z * rhs._x,
						_w * rhs._z + _x * rhs._y - 
						_y * rhs._x + _z * rhs._w);
					return tmp;
				}
				Quaternion& operator*=(const Quaternion& rhs)
				{
					_w = _w * rhs._w - _x * rhs._x - _y * rhs._y - _z * rhs._z;
					_x = _w * rhs._x + _x * rhs._w + _y * rhs._z - _z * rhs._y;
					_y = _w * rhs._y - _x * rhs._z + _y * rhs._w + _z * rhs._x;
					_z = _w * rhs._z + _x * rhs._y - _y * rhs._x + _z * rhs._w;
					return *this;
				}
				Quaternion operator*(const T rhs) const
				{
					Quaternion tmp(_w * rhs, _x * rhs, _y * rhs, _z * rhs);
					return tmp;
				}
				Quaternion& operator*=(const T rhs)
				{
					_w *= rhs; _x *= rhs; _y *= rhs; _z *= rhs;
					return *this;
				}
				friend Quaternion operator*(const T lhs, const Quaternion& rhs) 
				{
					Quaternion  tmp(rhs._w * lhs, rhs._x * lhs,
						rhs._y * lhs, rhs._z * lhs);
					return tmp;
				};
				Quaternion operator/(const Quaternion& rhs) const
				{
					Quaternion    tmp (~rhs / pow(rhs.getNorm(), 2) * *this);
					return tmp;
				}
				Quaternion& operator/=(const Quaternion& rhs)
				{
					*this = ~rhs / pow(rhs.getNorm(), 2) * *this;
					return *this;
				}
				Quaternion operator/(const T rhs) const
				{
					Quaternion    tmp (_w / rhs, _x / rhs, _y / rhs, _z / rhs);
					return tmp;
				}
				Quaternion operator~() const
				{ return conjugate(); }

				T getNorm() const 
				{ return (sqrt(_w * _w + _x * _x + _y * _y + _z * _z)); };
				void normalise()
				{
					float norm = getNorm();
					if (norm)
					{
						_w = _w / norm;
						_x = _x / norm;
						_y = _y / norm;
						_z = _z / norm;
					}
				}
				Quaternion conjugate() const
				{
					Quaternion tmp(_w, -_x, -_y, -_z);
					return tmp;
				}

				T getw() const {return _w;}
				T getx() const {return _x;}
				T gety() const {return _y;}
				T getz() const {return _z;}
				void setw(T in) {_w  = in;}
				void setx(T in) {_x  = in;}
				void sety(T in) {_y  = in;}
				void setz(T in) {_z  = in;}

				//
				// just the slerp function
				//
				static Quaternion<T>      slerp(const Quaternion<T>& q1, const Quaternion<T>& q2,
					const T step)
				{
					Quaternion    tmp;
					T         om = acos(q1.getw() * q2.getw() +
						q1.getx() * q2.getx() +
						q1.gety() * q2.gety() +
						q1.getz() * q2.getz());

					assert((step >= 0) && (step <= 1));
					if (sin(om) == 0)
						tmp = (q1 * sin((1 - step) * om) + q2 * sin(step * om)) / om;
					else
						tmp = (q1 * sin((1 - step) * om) + q2 * sin(step * om)) / sin(om);
					return tmp;
				}

	private:
		T	_w;
		T	_x;
		T	_y;
		T	_z;

	};
}

#endif // _TYPES_QUATERNION_HH_
