#ifndef _TYPES_VECT3D_HH_
#define _TYPES_VECT3D_HH_

#include <cmath>
#include "streamable.hh"
#include "toolance/fsqrt.hh"

// classe generique Vect3d<T>
// se construit Vect3d (T x, T y, T z)
//
// operations et fonctions implementes:
//
// implemente streamable (print et read)
// Vect3d + Vect3d
// Vect3d += Vect3d
// Vect3d - Vect3d
// Vect3d -= Vect3d
// Vect3d * Vect3d (produit vectoriel)
// Vect3d *= Vect3d (produit vectoriel)
// Vect3d mult Vect3d (produit membre a membre)
// get_norm (retourne la norme du Vect3d)
// normalise (normalise le Vect3d)
//
// les accesseurs sont getx()... et setx(T)...

namespace types 
{
	template <typename T>
	class Vect3d: public Streamable
	{
	public:
		Vect3d (): 
				_x (0),_y (0),_z (0) {}
				Vect3d (const T x, const T y, const T z): 
				_x (x),_y (y),_z (z) {}
				Vect3d (const Vect3d& rhs): 
				Streamable(), _x (rhs._x), _y (rhs._y), _z (rhs._z) {}
				Vect3d& operator= (const Vect3d& rhs) 
				{
					if (&rhs == this)	return *this;
					_x = rhs._x; _y = rhs._y; _z = rhs._z; return *this;
				}
				virtual ~Vect3d () {}

				void print(std::ostream& os) const 
				{ os << "(" << _x << " " << _y << " " << _z << ")"; }
				void read(std::istream& is) 
				{ is >> _x >> _y >> _z; }

				Vect3d<T> operator+ (const Vect3d& rhs) const
				{ 
					Vect3d<T> tmp (_x + rhs._x, _y + rhs._y, _z + rhs._z);
					return tmp;
				}
				Vect3d<T>& operator+= (const Vect3d& rhs)
				{
					_x += rhs._x; _y += rhs._y; _z += rhs._z;
					return *this;
				}
				Vect3d<T> operator- (const Vect3d& rhs) const
				{
					Vect3d<T> tmp (_x - rhs._x, _y - rhs._y, _z - rhs._z);
					return tmp;
				}
				Vect3d<T>& operator-= (const Vect3d& rhs)
				{
					_x -= rhs._x; _y -= rhs._y; _z -= rhs._z;
					return *this;
				}
				Vect3d<T> operator* (const Vect3d& rhs) const
				{
					Vect3d<T> tmp(_y * rhs._z - _z * rhs._y,
						_z * rhs._x - _x * rhs._z,
						_x * rhs._y - _y * rhs._x);
					return tmp;
				}

				Vect3d<T>& operator*= (const Vect3d& rhs)
				{
					_x = _x * rhs._y - _y * rhs._x;
					_y = _y * rhs._z - _z * rhs._y;
					_z = _z * rhs._x - _x * rhs._z;
					return *this;
				}

				Vect3d<T> operator* (const T& rhs) const
				{
					Vect3d<T> tmp (_x * rhs, _y * rhs, _z * rhs);
					return tmp;
				}


				Vect3d<T>& operator*= (const T& rhs)
				{
					_x = _x * rhs;
					_y = _y * rhs;
					_z = _z * rhs;
					return *this;
				}

				Vect3d<T> operator/ (const T& rhs) const
				{
					Vect3d<T> tmp (_x / rhs, _y / rhs, _z / rhs);
					return tmp;
				}

				Vect3d<T>& operator/= (const T& rhs)
				{
					_x = _x / rhs;
					_y = _y / rhs;
					_z = _z / rhs;
					return *this;
				}

				Vect3d<T> mult (const Vect3d& rhs) const
				{
					Vect3d<T> tmp(_x * rhs._x,
						_y * rhs._y,
						_z * rhs._z);
					return tmp;
				}

				T getNorm() const
				{ return (float)math::fsqrt(_x*_x + _y*_y + _z*_z); }

				Vect3d<T>& normalise ()
				{
					T norm = getNorm();
					if (norm)
					{ _x /= norm; _y /= norm; _z /= norm; }
					return *this;
				}

				T dot(const Vect3d<T>& rhs) const
				{
					return _x * rhs._x + _y * rhs._y + _z * rhs._z;
				}

				T getx() const {return _x;}
				T gety() const {return _y;}
				T getz() const {return _z;}
				void setx(T in) {_x  = in;}
				void sety(T in) {_y  = in;}
				void setz(T in) {_z  = in;}
				void set(T x, T y, T z)
				{ _x = x; _y = y; _z = z;}


				// Fait a - b et le met dans dst
				inline static void add(Vect3d<T>& dst, const Vect3d<T>& a, const Vect3d<T>& b)
				{
					dst._x = a._x + b._x;
					dst._y = a._y + b._y;
					dst._z = a._z + b._z;
				}


				// Fait a - b et le met dans dst
				inline static void sub(Vect3d<T>& dst, const Vect3d<T>& a, const Vect3d<T>& b)
				{
					dst._x = a._x - b._x;
					dst._y = a._y - b._y;
					dst._z = a._z - b._z;
				}

				// Fait a * b et le met dans dst
				inline static void mult(Vect3d<T>& dst, const Vect3d<T>& a, T b)
				{
					dst._x = a._x * b;
					dst._y = a._y * b;
					dst._z = a._z * b;
				}

				// Fait a + (b * c) et le met dans dst
				inline static void addmult(Vect3d<T>& dst, const Vect3d<T>& a, const Vect3d<T>& b, T c)
				{
					dst._x = a._x + b._x * c;
					dst._y = a._y + b._y * c;
					dst._z = a._z + b._z * c;
				}

				// Fait a - (b * c) et le met dans dst
				inline static void submult(Vect3d<T>& dst, const Vect3d<T>& a, const Vect3d<T>& b, T c)
				{
					dst._x = a._x - b._x * c;
					dst._y = a._y - b._y * c;
					dst._z = a._z - b._z * c;
				}



	private:
		T	_x;
		T	_y;
		T	_z;
	};
}

#endif // _TYPES_VECT3D_HH_
